import { assertNever as assertTypeIsHandled } from "../utils/lang";

export type AudioSampleId = string;

export type AudioAssetKind = "music" | "effect";

export type AudioAssetPartialInformation = {
  name: string;
  kind: AudioAssetKind;
};

export type AudioAssets = {
  snippets: {
    [id: string]: {
      kind: AudioAssetKind;
      media: HTMLAudioElement;
    };
  };
};

/**
 * @param {import("./types").AudioAssets} audioAssets The sound snippets available in-game
 */
export class AudioEngine {
  private context: AudioContext;
  private musicGain: GainNode;
  private effectGain: GainNode;
  private registeredTracks: Record<
    AudioSampleId,
    {
      audio: HTMLMediaElement;
      gain: GainNode;
      track: MediaElementAudioSourceNode;
    }
  >;

  constructor(audioAssets: AudioAssets) {
    this.context = new window.AudioContext();
    this.musicGain = this.context.createGain();
    this.musicGain.gain.value = 1.0;
    this.effectGain = this.context.createGain();
    this.effectGain.gain.value = 1.0;
    this.registeredTracks = {};

    for (const assetId in audioAssets.snippets) {
      const asset = audioAssets.snippets[assetId];
      if (this.registeredTracks[assetId])
        throw new Error(`Cannot register the audio asset ${assetId} twice`);
      const track = this.context.createMediaElementSource(asset.media);
      const gain = this.context.createGain();
      let connectedNode = track.connect(gain);
      if (asset.kind === "music")
        connectedNode = connectedNode.connect(this.musicGain);
      else if (asset.kind === "effect")
        connectedNode = connectedNode.connect(this.effectGain);
      connectedNode.connect(this.context.destination);
      gain.gain.value = 0.0;
      this.registeredTracks[assetId] = {
        audio: asset.media,
        gain: gain,
        track: track,
      };
    }
  }

  public setVolume(kind: AudioAssetKind, value: number) {
    if (kind === "effect") {
      this.effectGain.gain.setTargetAtTime(
        value,
        this.context.currentTime,
        0.01,
      );
      return;
    }
    if (kind === "music") {
      this.musicGain.gain.setTargetAtTime(
        value,
        this.context.currentTime,
        0.01,
      );
      return;
    }
    assertTypeIsHandled(kind, "Audio track kind");
  }

  public startLoop(identifier: AudioSampleId) {
    const data = this.registeredTracks[identifier];
    if (!data)
      throw new Error(
        `Cannot enable audio for unknown identifier ${identifier}`,
      );
    if (data.audio.paused) {
      try {
        data.audio.loop = true;
        data.audio.currentTime = 0;
        data.audio.play();
      } catch (e) {}
    }
  }

  public stopLoop(identifier: AudioSampleId) {
    const data = this.registeredTracks[identifier];
    if (!data)
      throw new Error(
        `Cannot enable audio for unknown identifier ${identifier}`,
      );
    if (!data.audio.paused) {
      try {
        data.audio.pause();
      } catch (e) {}
    }
  }

  public loopDisable(identifier: AudioSampleId) {
    const data = this.registeredTracks[identifier];
    if (!data)
      throw new Error(
        `Cannot enable audio for unknown identifier ${identifier}`,
      );
    try {
      data.gain.gain.setTargetAtTime(0.0, this.context.currentTime, 0.2);
    } catch (e) {}
  }

  public loopEnable(identifier: AudioSampleId) {
    const data = this.registeredTracks[identifier];
    if (!data)
      throw new Error(
        `Cannot enable audio for unknown identifier ${identifier}`,
      );
    try {
      data.gain.gain.setTargetAtTime(1.0, this.context.currentTime, 0.2);
    } catch (e) {}
  }

  public playOnce(identifier: AudioSampleId) {
    const data = this.registeredTracks[identifier];
    if (!data)
      throw new Error(
        `Cannot enable audio for unknown identifier ${identifier}`,
      );
    try {
      data.gain.gain.value = 1.0;
      data.audio.loop = false;
      data.audio.currentTime = 0;
      data.audio.play();
    } catch (e) {}
  }

  public stopAll() {
    for (const sampleId in this.registeredTracks) {
      try {
        this.registeredTracks[sampleId].audio.pause();
      } catch (e) {}
    }
  }
}
