import { PIXI } from "../pixi";
import { createOnScreenControlsIfNeeded } from "./on-screen";

export type CurrentControlData = {
  up: boolean;
  down: boolean;
  left: boolean;
  right: boolean;
  action1: boolean;
  action2: boolean;
  back: boolean;
};

export type PointerControlData = {
  x: number;
  y: number;
  pressed: boolean;
  pressed_on_frame: boolean;
  held: boolean;
  updateButtonStates(): void;
};

export type ControlMappings = {
  [k: string]: keyof CurrentControlData;
};

export type ControlData = {
  keys: CurrentControlData;
  pointer: PointerControlData;
  mappings: ControlMappings;
};

export function createControls(application: PIXI.Application): ControlData {
  const controls: ControlData = {
    keys: {
      up: false,
      down: false,
      left: false,
      right: false,
      action1: false,
      action2: false,
      back: false,
    },
    pointer: {
      x: 0,
      y: 0,
      pressed: false,
      pressed_on_frame: false,
      held: false,
      updateButtonStates: updateMouseButtonStates,
    },
    mappings: {
      ArrowUp: "up",
      ArrowDown: "down",
      ArrowLeft: "left",
      ArrowRight: "right",
      w: "up",
      a: "left",
      s: "down",
      d: "right",
      z: "up",
      q: "left",
      " ": "action1",
      Shift: "action2",
      PageUp: "back",
      o: "action1",
      p: "action2",
      Backspace: "back",
    },
  };

  function updateMouseButtonStates() {
    if (controls.pointer.pressed) {
      if (!controls.pointer.held) {
        controls.pointer.held = true;
        controls.pointer.pressed_on_frame = true;
      } else {
        controls.pointer.pressed_on_frame = false;
      }
    } else {
      controls.pointer.pressed_on_frame = false;
      controls.pointer.held = false;
    }
  }

  document.addEventListener("touchstart", () => {
    createOnScreenControlsIfNeeded(controls);
  });

  document.addEventListener("keydown", (ev) => {
    if (controls.mappings[ev.key]) {
      controls.keys[controls.mappings[ev.key]] = true;
    }
  });
  document.addEventListener("keyup", (ev) => {
    if (controls.mappings[ev.key]) {
      controls.keys[controls.mappings[ev.key]] = false;
    }
  });

  application.stage.interactive = true;
  application.stage.on("mousedown", (event) => {
    controls.pointer.x = event.global.x;
    controls.pointer.y = event.global.y;
    if (event.button === 0) controls.pointer.pressed = true;
  });

  application.stage.on("mousemove", (event) => {
    controls.pointer.x = event.global.x;
    controls.pointer.y = event.global.y;
  });

  application.stage.on("mouseup", (event) => {
    controls.pointer.x = event.global.x;
    controls.pointer.y = event.global.y;
    if (event.button === 0) controls.pointer.pressed = false;
  });

  return controls;
}
