import { ControlData } from ".";
import "./on-screen-style.css";

type OnScreenControlElements = {
  container: HTMLElement;
  up: HTMLElement;
  down: HTMLElement;
  left: HTMLElement;
  right: HTMLElement;
  action1: HTMLElement;
  action2: HTMLElement;
  back: HTMLElement;
  hide: HTMLElement;
};

function getSamplePointsFromEllipse(
  x: number,
  y: number,
  skewedXRadius: number,
  skewedYRadius: number,
  rAngle: number,
): Array<{ x: number; y: number }> {
  const cos = Math.cos(rAngle);
  const sin = Math.sin(rAngle);
  const SAMPLE_STEP_SIZE = 5;
  const samples = [{ x, y }];
  for (
    let skewedDeltaX = 0;
    skewedDeltaX < skewedXRadius;
    skewedDeltaX += SAMPLE_STEP_SIZE
  ) {
    for (
      let skewedDeltaY = 0;
      skewedDeltaY < skewedYRadius;
      skewedDeltaY += SAMPLE_STEP_SIZE
    ) {
      if (skewedDeltaX === 0 && skewedDeltaY === 0) {
        continue;
      }
      const adx = skewedDeltaX * cos + skewedDeltaY * sin;
      const ady = skewedDeltaX * sin + skewedDeltaY * cos;
      samples.push({ x: x - adx, y: y - ady });
      samples.push({ x: x - adx, y: y + ady });
      samples.push({ x: x + adx, y: y - ady });
      samples.push({ x: x + adx, y: y + ady });
    }
  }

  return samples;
}

function createOnScreenControlElements(): OnScreenControlElements {
  const container = document.createElement("div");
  container.className = "onscreen-controls";

  const directionPad = document.createElement("nav");
  directionPad.className = "onscreen-direction-pad";
  const up = document.createElement("a");
  up.className = "up";
  directionPad.appendChild(up);
  const right = document.createElement("a");
  right.className = "right";
  directionPad.appendChild(right);
  const down = document.createElement("a");
  down.className = "down";
  directionPad.appendChild(down);
  const left = document.createElement("a");
  left.className = "left";
  directionPad.appendChild(left);
  container.appendChild(directionPad);

  const actionPad = document.createElement("div");
  actionPad.className = "onscreen-action-pad";
  const fullScreen = document.createElement("button");
  fullScreen.textContent = !document.fullscreenElement
    ? "Full screen"
    : "Exit full screen";
  fullScreen.addEventListener("click", () => {
    if (!document.fullscreenElement) {
      document.documentElement.requestFullscreen();
      fullScreen.textContent = "Exit full screen";
    } else if (document.exitFullscreen) {
      document.exitFullscreen();
      fullScreen.textContent = "Full screen";
    }
  });
  actionPad.appendChild(fullScreen);
  const hide = document.createElement("div");
  hide.className = "hide";
  hide.textContent = "Hide controls";
  actionPad.appendChild(hide);
  const action1 = document.createElement("div");
  action1.className = "action";
  action1.textContent = "Action";
  actionPad.appendChild(action1);
  const action2 = document.createElement("div");
  action2.className = "action";
  action2.textContent = "Sprint";
  actionPad.appendChild(action2);
  const back = document.createElement("div");
  back.className = "hide";
  back.textContent = "Back";
  actionPad.appendChild(back);
  container.appendChild(actionPad);
  return {
    container,
    up,
    down,
    left,
    right,
    action1,
    action2,
    back,
    hide,
  };
}

const keys = [
  "up",
  "down",
  "left",
  "right",
  "action1",
  "action2",
  "back",
] as const;

let hasOnScreenControls = false;
export function createOnScreenControlsIfNeeded(controlData: ControlData) {
  if (hasOnScreenControls) return;
  hasOnScreenControls = true;
  const body = document.querySelector("body");
  if (!body) return;
  const unsubscribers: Array<() => void> = [];
  const elements = createOnScreenControlElements();
  body.appendChild(elements.container);
  const checkTouchStates = (ev: TouchEvent) => {
    let touchedItems = [];
    for (let touchIndex = 0; touchIndex < ev.touches.length; touchIndex++) {
      const touch = ev.touches.item(touchIndex);
      if (!touch) continue;
      const samples = getSamplePointsFromEllipse(
        touch.pageX,
        touch.pageY,
        15,
        15,
        touch.rotationAngle,
      );
      for (const sample of samples) {
        touchedItems.push(document.elementFromPoint(sample.x, sample.y));
      }
    }
    for (const key of keys) {
      const isPressed = touchedItems.some((t) => t === elements[key]);
      controlData.keys[key] = isPressed;
      if (isPressed) {
        elements[key].classList.add("pressed");
      } else {
        elements[key].classList.remove("pressed");
      }
    }
  };
  body.addEventListener("touchstart", checkTouchStates);
  body.addEventListener("touchmove", checkTouchStates);
  body.addEventListener("touchend", checkTouchStates);
  body.addEventListener("touchcancel", checkTouchStates);
  unsubscribers.push(() => {
    body.removeEventListener("touchstart", checkTouchStates);
    body.removeEventListener("touchmove", checkTouchStates);
    body.removeEventListener("touchend", checkTouchStates);
    body.removeEventListener("touchcancel", checkTouchStates);
  });
  for (const key of keys) {
    const press = () => (controlData.keys[key] = true);
    const release = () => (controlData.keys[key] = false);
    elements[key].addEventListener("mousedown", press);
    elements[key].addEventListener("mouseup", release);
    unsubscribers.push(() => {
      elements[key].removeEventListener("mousedown", press);
      elements[key].removeEventListener("mouseup", release);
    });
  }
  const hideOnScreenControls = () => removeOnScreenControls();
  elements.hide.addEventListener("mousedown", hideOnScreenControls);
  elements.hide.addEventListener("touchstart", hideOnScreenControls);
  unsubscribers.push(() => {
    elements.hide.removeEventListener("mousedown", hideOnScreenControls);
    elements.hide.removeEventListener("touchstart", hideOnScreenControls);
  });

  function removeOnScreenControls() {
    for (const unsubscriber of unsubscribers) {
      unsubscriber();
    }
    elements.container.remove();
    hasOnScreenControls = false;
  }
}
