import { AudioEngine } from "./audio";
import { ControlData } from "./controls";
import { EntityType } from "./entities";
import { ImageRenderingData } from "./images";
import { PIXI } from "./pixi";

export type GameData = {
  entities: EntityType[];
  stats: {
    gatheredPlayerMoney: number;
    gatheredTownieMoney: number;
    gatheredEnemyMoney: number;
    stolenEnemyMoney: number;
    retrievedEnemyMoney: number;
  };
};

export type ControllerSystemMethod = (game: Game, elapsedTicks: number) => void;

export type SystemMethod = (game: Game) => void;

export class Game {
  private controllerSystems: ControllerSystemMethod[];
  private rendererSystems: SystemMethod[];

  public fps: number;
  public readonly audio: AudioEngine;
  public soundQueue: Array<{ sound: "error" | "click" }>;
  public readonly controls: ControlData;
  public readonly application: PIXI.Application;
  public readonly mapContainer: PIXI.Container;
  public readonly uiContainer: PIXI.Container;
  public readonly images: ImageRenderingData;
  public readonly gameData: GameData;

  constructor(
    baseGameData: GameData,
    systems: {
      controller: ControllerSystemMethod[];
      render: SystemMethod[];
    },
    engines: {
      application: PIXI.Application;
      audio: AudioEngine;
      controls: ControlData;
      images: ImageRenderingData;
    },
  ) {
    this.fps = 0;
    this.application = engines.application;
    this.mapContainer = new PIXI.Container();
    this.application.stage.addChild(this.mapContainer);
    this.mapContainer.sortableChildren = true;
    this.uiContainer = new PIXI.Container();
    this.application.stage.addChild(this.uiContainer);
    this.images = engines.images;
    this.audio = engines.audio;
    this.controls = engines.controls;
    this.soundQueue = [];
    this.controllerSystems = systems.controller;
    this.rendererSystems = systems.render;
    this.gameData = baseGameData;
  }

  update(elapsedTicks: number) {
    this.controls.pointer.updateButtonStates();

    for (const system of this.controllerSystems) {
      system(this, elapsedTicks);
    }
  }

  render(fps: number) {
    // Update the FPS counter
    this.fps = fps;

    // Add the renderers as needed
    for (const entity of this.gameData.entities) {
      if (entity.mapRenderer && !entity.mapRenderer.isRendering) {
        this.mapContainer.addChild(entity.mapRenderer.pixiObject);
        entity.mapRenderer.isRendering = true;
      }
      if (entity.uiRenderer && !entity.uiRenderer.isRendering) {
        this.uiContainer.addChild(entity.uiRenderer.pixiObject);
        entity.uiRenderer.isRendering = true;
      }
    }

    for (const system of this.rendererSystems) {
      system(this);
    }
  }

  extraEvents() {
    for (const event of this.soundQueue) {
      this.audio.playOnce(event.sound);
    }

    this.soundQueue = [];

    // Remove the entities to delete as needed
    const entitiesToDelete = this.gameData.entities.filter(
      (e) => e.deletionFlag,
    );
    for (const entity of entitiesToDelete) {
      if (entity.mapRenderer) entity.mapRenderer.pixiObject.removeFromParent();
      if (entity.uiRenderer) entity.uiRenderer.pixiObject.removeFromParent();
    }
    this.gameData.entities = this.gameData.entities.filter(
      (e) => !e.deletionFlag,
    );
  }
}
