import { AudioEngine } from "./audio";
import { createControls } from "./controls";
import { Game } from "./engine";
import { ImageRenderingData, loadAllSpritesheets } from "./images";
import { PIXI } from "./pixi";
import { createTitleScene } from "./scene/title";
import "./style.css";
import { controllerAIAllySystem } from "./systems/controller-ai-ally-system";
import { controllerAIDeliverySystem } from "./systems/controller-ai-delivery-system";
import { controllerAIEnemySystem } from "./systems/controller-ai-enemy-system";
import { controllerAIGuardSystem } from "./systems/controller-ai-guard-system";
import { controllerControlsPlayerSystem } from "./systems/controller-controls-player-system";
import { controllerControlsTitleScreenSystem } from "./systems/controller-controls-title-screen-system";
import { controllerCharacterMovementSystem } from "./systems/controller-map-character-movement-system";
import { controllerMapPathingSystem } from "./systems/controller-map-pathing-system";
import { controllerSpawnerAllySystem } from "./systems/controller-spawner-ally-system";
import { controllerSpawnerDeliverySystem } from "./systems/controller-spawner-delivery-system";
import { controllerSpawnerEnemySystem } from "./systems/controller-spawner-enemy-system";
import { controllerSpawnerGuardSystem } from "./systems/controller-spawner-guard-system";
import { controllerSpriteCaptureSystem } from "./systems/controller-sprite-capture-system";
import { controllerSpriteCharacterAnimationSystem } from "./systems/controller-sprite-character-animation-system";
import { controllerSpriteCharacterInventorySystem } from "./systems/controller-sprite-character-inventory-system";
import { controllerUIPlayerSystem } from "./systems/controller-ui-player-system";
import { rendererCharacterSpriteUpdaterSystem } from "./systems/renderer-character-sprite-updater-system";
import { rendererMapCameraPositionSystem } from "./systems/renderer-map-camera-position-system";
import { rendererMapPositionSystem } from "./systems/renderer-map-position-system";
import { rendererUIPlayerSystem } from "./systems/renderer-ui-player-system";

const MS_BETWEEN_TICKS = 1000 / 20; /* 20 ticks per second*/
const MAX_MS_BETWEEN_FRAMES = 1000 / 60; /* capped at 60 FPS */

(async () => {
  const appElement = document.getElementById("app");
  if (!appElement) return;

  const images = await loadAllSpritesheets();

  const app = new PIXI.Application();
  await app.init({ width: 800, height: 600 });
  appElement.appendChild(app.canvas);
  startMainMenu(app, images);
})();

function startMainMenu(
  application: PIXI.Application,
  images: ImageRenderingData,
) {
  const game = new Game(
    {
      entities: [],
      stats: {
        gatheredTownieMoney: 0,
        gatheredEnemyMoney: 0,
        gatheredPlayerMoney: 0,
        retrievedEnemyMoney: 0,
        stolenEnemyMoney: 0,
      },
    },
    {
      controller: [
        controllerSpawnerAllySystem(),
        controllerSpawnerDeliverySystem(),
        controllerSpawnerEnemySystem(),
        controllerSpawnerGuardSystem(),
        controllerAIAllySystem(),
        controllerAIDeliverySystem(),
        controllerAIEnemySystem(),
        controllerAIGuardSystem(),
        controllerControlsPlayerSystem(),
        controllerControlsTitleScreenSystem(),
        controllerCharacterMovementSystem(),
        controllerSpriteCharacterInventorySystem(),
        controllerMapPathingSystem(),
        controllerSpriteCharacterAnimationSystem(),
        controllerUIPlayerSystem(),
        controllerSpriteCaptureSystem(),
      ],
      render: [
        rendererUIPlayerSystem(),
        rendererCharacterSpriteUpdaterSystem(),
        rendererMapPositionSystem(),
        rendererMapCameraPositionSystem(),
      ],
    },
    {
      application,
      audio: new AudioEngine({
        snippets: {},
      }),
      controls: createControls(application),
      images,
    },
  );

  createTitleScene(game);

  let lastTick = Date.now();
  let lastRender = Date.now();
  function mainMenuLoop() {
    const currentLoop = Date.now();
    if (currentLoop >= lastTick + MS_BETWEEN_TICKS) {
      const elapsedMsBetweenNowAndThen = currentLoop - lastTick;
      const remainingMs = elapsedMsBetweenNowAndThen % MS_BETWEEN_TICKS;
      const elapsedTicks =
        (elapsedMsBetweenNowAndThen - remainingMs) / MS_BETWEEN_TICKS;
      game.update(elapsedTicks);
      lastTick = currentLoop - remainingMs;
    }
    if (currentLoop >= lastRender + MAX_MS_BETWEEN_FRAMES) {
      const fps = 1000 / (currentLoop - lastRender);
      game.render(fps);
      lastRender = currentLoop;
    }
    game.extraEvents();
  }
  application.ticker.add(mainMenuLoop);
}
