import { distanceBetween } from "../utils/analyzer";
import {
  getProbabilityResult,
  getRandomArrayItem,
  getRandomIntegerInRange,
} from "../utils/random";

export const SMALL_TILE_SIZE = 20;
export const LARGE_TILE_SIZE = 42;

const params = {
  borderX: 4,
  borderY: 4,
  tileSize: LARGE_TILE_SIZE,
  sizeX: 60,
  sizeY: 60,
  maxTownAttempts: 10,
  maxCampAttempts: 20,
  minTownDistance: 16,
  minCampToTownDistance: 7,
  minCampBetweenDistance: 7,
  maxTreeAttempts: 100,
  townsToForceConnect: 2,
} as const;

export type TownData = {
  x: number;
  y: number;
  image: "red" | "blue";
};

export type CampData = {
  x: number;
  y: number;
  image: "tent" | "fire";
};

export type RoadData = {
  x: number;
  y: number;
  north: boolean;
  south: boolean;
  east: boolean;
  west: boolean;
};

export type TreeData = {
  x: number;
  y: number;
  tile: "1" | "2";
};

export type TownGraph = Array<[TownData, TownData]>;

export type MapData = {
  towns: TownData[];
  graph: TownGraph;
  camps: CampData[];
  roads: RoadData[];
  trees: TreeData[];
  tileSize: number;
  sizeX: number;
  sizeY: number;
};

function getRandomTowns() {
  const towns: TownData[] = [];
  const townToTownDistances: Map<TownData, Map<TownData, number>> = new Map();
  function getDistanceMap(town: TownData): Map<TownData, number> {
    let map = townToTownDistances.get(town);
    if (!map) {
      map = new Map();
      townToTownDistances.set(town, map);
    }
    return map;
  }
  for (let i = 0; i < params.maxTownAttempts; ++i) {
    const x = getRandomIntegerInRange(
      params.borderX,
      params.sizeX - params.borderX,
    );
    const y = getRandomIntegerInRange(
      params.borderY,
      params.sizeY - params.borderY,
    );
    if (
      towns.some(
        (town) => distanceBetween({ x, y }, town) < params.minTownDistance,
      )
    )
      continue;
    const newTown: TownData = {
      x,
      y,
      image: getRandomArrayItem(["red", "blue"]),
    };
    towns.push(newTown);

    // Update distances map
    const newTownMap = getDistanceMap(newTown);
    for (const otherTown of towns) {
      const distance = distanceBetween(newTown, otherTown);
      newTownMap.set(otherTown, distance);
      getDistanceMap(otherTown).set(newTown, distance);
    }
  }
  return { towns, townToTownDistances };
}

function getTownGraphWithKruskal(
  towns: ReadonlyArray<TownData>,
  townToTownDistances: ReadonlyMap<TownData, ReadonlyMap<TownData, number>>,
): TownGraph {
  // algorithm Kruskal(G) is
  //   for each (u, v) in G.E ordered by weight(u, v), increasing do
  //       if FIND-SET(u) ≠ FIND-SET(v) then
  //           F:= F ∪ {(u, v)} ∪ {(v, u)}
  //           UNION(FIND-SET(u), FIND-SET(v))
  //   return F
  const pairs = towns
    .flatMap((t) => [
      ...towns.map(
        (u) => [t, u, townToTownDistances.get(t)?.get(u) || 0] as const,
      ),
    ])
    .filter(([p1, p2]) => p1 !== p2);
  pairs.sort(([_1, _2, d1], [_3, _4, d2]) => d1 - d2);
  const townsWithIndirectPathBetween: Map<
    TownData,
    Map<TownData, true>
  > = new Map();
  function getTownPaths(town: TownData): Map<TownData, true> {
    let townPaths = townsWithIndirectPathBetween.get(town);
    if (!townPaths) {
      townPaths = new Map();
      townsWithIndirectPathBetween.set(town, townPaths);
    }
    return townPaths;
  }

  function addIndirectPaths(source: TownData, target: TownData) {
    const sourcePaths = getTownPaths(source);
    const targetPaths = getTownPaths(target);
    sourcePaths.set(target, true);
    targetPaths.set(source, true);
    for (const connectedToSource of sourcePaths.keys()) {
      const connectedToSourcePaths = getTownPaths(connectedToSource);
      targetPaths.set(connectedToSource, true);
      connectedToSourcePaths.set(target, true);
      for (const connectedToTarget of sourcePaths.keys()) {
        const connectedToTargetPaths = getTownPaths(connectedToTarget);
        connectedToTargetPaths.set(source, true);
        connectedToTargetPaths.set(connectedToSource, true);
        connectedToSourcePaths.set(connectedToTarget, true);
      }
    }
  }

  let forceConnected = 0;
  const graph: TownGraph = [];
  for (const [town1, town2, _distance] of pairs) {
    if (getTownPaths(town1).get(town2)) {
      const alreadyConnected = graph.some(
        ([t1, t2]) =>
          (t1 === town1 && t2 === town2) || (t2 === town1 && t1 === town2),
      );
      if (alreadyConnected) continue;
      if (forceConnected >= params.townsToForceConnect) {
        continue;
      }
      forceConnected++;
    }
    graph.push([town1, town2]);
    addIndirectPaths(town1, town2);
    addIndirectPaths(town2, town1);
  }
  return graph;
}

function createRoadsFromTownGraph(towns: TownData[], graph: TownGraph) {
  const roads: RoadData[] = [];

  // Create town squares (the two bottom tiles of towns)
  for (const town of towns) {
    roads.push({
      x: town.x,
      y: town.y + 1,
      north: true,
      south: false,
      east: true,
      west: false,
    });
    roads.push({
      x: town.x + 1,
      y: town.y + 1,
      north: true,
      south: false,
      east: false,
      west: true,
    });
  }

  // Turn graph lines into roads
  for (const [town1, town2] of graph) {
    let startX = town1.x;
    if (town1.x < town2.x) startX++;
    let startY = town1.y + 1;
    let endX = town2.x;
    if (town2.x < town1.x) endX++;
    let endY = town2.y + 1;
    let lastDirection: "n" | "s" | "e" | "w" = "n";
    while (true) {
      const distanceX = Math.abs(endX - startX);
      const distanceY = Math.abs(endY - startY);
      if (distanceX < distanceY) {
        // Move on the Y direction.
        const north = lastDirection === "n" || endY < startY;
        const south = lastDirection === "s" || endY > startY;
        const east = lastDirection === "e";
        const west = lastDirection === "w";
        roads.push({ x: startX, y: startY, north, south, east, west });
        lastDirection = endY > startY ? "n" : "s";
        if (endY > startY) startY++;
        else startY--;
        continue;
      }
      // Move on the X direction.
      const north = lastDirection === "n";
      const south = lastDirection === "s";
      const east = lastDirection === "e" || endX > startX;
      const west = lastDirection === "w" || endX < startX;
      roads.push({ x: startX, y: startY, north, south, east, west });
      lastDirection = endX > startX ? "w" : "e";
      if (startX === endX && startY === endY) break;
      if (endX > startX) startX++;
      else startX--;
    }
  }

  return roads;
}

function mergeRoads(roads: RoadData[]): RoadData[] {
  // Merge any road tiles that are actually the same
  const resultSet: Map<string, RoadData> = new Map();
  for (const road of roads) {
    const existing = resultSet.get(`${road.x},${road.y}`);
    if (existing) {
      existing.north = existing.north || road.north;
      existing.south = existing.south || road.south;
      existing.east = existing.east || road.east;
      existing.west = existing.west || road.west;
      continue;
    }
    resultSet.set(`${road.x},${road.y}`, road);
  }

  return [...resultSet.values()];
}

function createCamps(towns: TownData[], roads: RoadData[]): CampData[] {
  const possibleTargets = roads.filter(
    (r) =>
      !towns.some((t) => distanceBetween(r, t) < params.minCampToTownDistance),
  );
  if (possibleTargets.length === 0) return [];
  const camps: CampData[] = [];
  for (let i = 0; i < params.maxCampAttempts; ++i) {
    const { x, y: rawY } = getRandomArrayItem(possibleTargets);
    const y = rawY - 1;
    if (
      camps.some(
        (camp) =>
          distanceBetween({ x, y }, camp) < params.minCampBetweenDistance,
      )
    )
      continue;
    camps.push({ x, y, image: getRandomArrayItem(["tent", "fire"]) });
  }
  return camps;
}

function createTrees(towns: TownData[], roads: RoadData[], camps: CampData[]) {
  const usedTiles = ([] as Array<[number, number]>)
    .concat(roads.map((t) => [t.x, t.y]))
    .concat(
      towns.flatMap((t) => [
        [t.x, t.y],
        [t.x + 1, t.y],
        [t.x, t.y + 1],
        [t.x + 1, t.y + 1],
      ]),
    )
    .concat(
      camps.flatMap((t) => [
        [t.x, t.y],
        [t.x + 1, t.y],
        [t.x, t.y + 1],
        [t.x + 1, t.y + 1],
      ]),
    );
  const trees: TreeData[] = [];

  // Add some trees around the edge of the map
  for (let x = 0; x < params.borderX; ++x) {
    for (let y = 0; y < params.sizeY; ++y) {
      if (getProbabilityResult(0.6))
        trees.push({ x, y, tile: getRandomArrayItem(["1", "2"]) });
      if (getProbabilityResult(0.6))
        trees.push({
          x: params.sizeX - 1 - x,
          y,
          tile: getRandomArrayItem(["1", "2"]),
        });
    }
  }
  for (let x = 3; x < params.sizeX - params.borderX; ++x) {
    for (let y = 0; y < params.borderY; ++y) {
      if (getProbabilityResult(0.6))
        trees.push({ x, y, tile: getRandomArrayItem(["1", "2"]) });
      if (getProbabilityResult(0.6))
        trees.push({
          x,
          y: params.sizeY - 1 - y,
          tile: getRandomArrayItem(["1", "2"]),
        });
    }
  }

  // Add a bunch of random trees in the map itself
  let usedSpots = new Set(usedTiles.map(([x, y]) => `${x},${y}`));
  for (let i = 0; i < params.maxTreeAttempts; ++i) {
    const x = getRandomIntegerInRange(
      params.borderX,
      params.sizeX - params.borderX,
    );
    const y = getRandomIntegerInRange(
      params.borderY,
      params.sizeY - params.borderY,
    );
    if (usedSpots.has(`${x},${y}`)) continue;
    usedSpots.add(`${x},${y}`);
    trees.push({ x, y: y - 1, tile: getRandomArrayItem(["1", "2"]) });
  }

  return trees;
}

export function createRandomMap(): MapData {
  // Create a set of random towns
  const { towns, townToTownDistances } = getRandomTowns();

  // Ccreate a town graph using Kruskal's algorithm
  const graph = getTownGraphWithKruskal(towns, townToTownDistances);

  // Connect the towns with roads
  const roads = createRoadsFromTownGraph(towns, graph);

  // Add some camps
  const camps = createCamps(towns, roads);

  // Add some trees
  const trees: TreeData[] = createTrees(towns, roads, camps);

  return {
    towns,
    graph,
    roads: mergeRoads(roads),
    camps,
    trees,
    tileSize: params.tileSize,
    sizeX: params.sizeX,
    sizeY: params.sizeY,
  };
}
