import { ControllerSystemMethod, Game } from "../engine";
import { EntityType } from "../entities";
import {
  distanceBetween,
  getDirectionFromTargetToTarget,
  getIsReachableMethod,
} from "../utils/analyzer";

export function controllerAIEnemySystem(): ControllerSystemMethod {
  return (game, elapsedTicks) => {
    for (const entity of game.gameData.entities) {
      if (!entity.aiEnemy) continue;
      checkStruggling(elapsedTicks, entity);
      checkTargetBeingCaptured(game, elapsedTicks, entity);
      checkTargetToChase(game, entity);
      checkTargetCampArrival(game, entity);
    }
  };
}

function getCharacterToChase(
  game: Game,
  position: { x: number; y: number },
  direction: "n" | "s" | "e" | "w",
  maxChaseDistance: number,
): { character: EntityType; distance: number } | undefined {
  const reachable = getIsReachableMethod(
    position,
    direction,
    maxChaseDistance,
    1.5,
    0.5,
  );

  const capturePreventions = game.gameData.entities
    .filter((e) => e.mapPosition && e.capturePrevention)
    .map((e) => ({
      x: e.mapPosition!.x,
      y: e.mapPosition!.y,
      radius: e.capturePrevention!.radius,
    }));

  const actualTargets: EntityType[] = [];
  for (const target of game.gameData.entities) {
    if (!target.capturable) continue;
    if (!target.capturable.isCapturableByBandits) continue;
    if (!target.inventory) continue;
    if (target.inventory.money <= 0 && target.inventory.clothes === "underwear")
      continue;
    if (!target.mapPosition) continue;
    const position = target.mapPosition;
    const isNearCaptureBlocker = capturePreventions.some(
      (t) => distanceBetween(position, t) < t.radius,
    );
    if (isNearCaptureBlocker) continue;
    if (direction === "n" && position.y < position.y) continue;
    if (direction === "s" && position.y > position.y) continue;
    if (direction === "w" && position.x < position.x) continue;
    if (direction === "e" && position.x > position.x) continue;
    actualTargets.push(target);
  }

  let closest: { character: EntityType; distance: number } | undefined;
  for (const target of actualTargets) {
    if (!target.mapPosition) continue;
    if (!reachable(target.mapPosition)) continue;
    const d = distanceBetween(target.mapPosition, position);
    if (d > maxChaseDistance) continue;
    if (closest && closest.distance < d) continue;
    closest = { character: target, distance: d };
  }
  return closest;
}

function checkStruggling(elapsedTicks: number, entity: EntityType) {
  if (!entity.aiEnemy) return;
  if (!entity.capturable) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPathing) return;
  if (!entity.characterSpriteRenderUpdater) return;
  const ai = entity.aiEnemy;
  const capturable = entity.capturable;
  const movement = entity.mapMovement;
  const pathing = entity.mapPathing;
  const renderer = entity.characterSpriteRenderUpdater;

  movement.enabled = false;
  pathing.enabled = false;

  if (capturable.state !== "bound") {
    renderer.struggleDataPercent = undefined;
    return;
  }

  capturable.remainingTightness -= elapsedTicks * ai.struggleStrength;
  renderer.struggleDataPercent =
    capturable.remainingTightness / capturable.totalTightness;
  if (capturable.remainingTightness < 0) {
    capturable.state = "free";
    movement.enabled = true;
    pathing.enabled = true;
  }
}

function checkTargetBeingCaptured(
  game: Game,
  elapsedTicks: number,
  entity: EntityType,
) {
  if (!entity.aiEnemy) return;
  if (!entity.inventory) return;
  if (!entity.mapPosition) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPathing) return;
  if (!entity.characterSpriteRenderUpdater) return;
  const ai = entity.aiEnemy;
  const movement = entity.mapMovement;
  const pathing = entity.mapPathing;
  const inventory = entity.inventory;
  const renderer = entity.characterSpriteRenderUpdater;
  if (!ai.targetBeingCaptured) return;

  movement.enabled = false;
  pathing.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;

  const targetId = ai.targetBeingCaptured;
  const target = game.gameData.entities.find((e) => e.id === targetId);
  if (!target) {
    ai.targetBeingCaptured = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  if (!target.capturable) {
    ai.targetBeingCaptured = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  if (entity.capturable && entity.capturable.state !== "free") {
    if (target.capturable.state === "grabbed") target.capturable.state = "free";
    if (target.mapMovement) target.mapMovement.enabled = true;
    ai.targetBeingCaptured = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  if (target.capturable.state === "grabbed") {
    if (ai.remainingCaptureTime > 0) {
      ai.remainingCaptureTime -= elapsedTicks;
      return;
    }
    target.capturable.state = "bound";
    target.capturable.totalTightness = ai.captureStrength;
    target.capturable.remainingTightness = ai.captureStrength;
  }

  if (target.inventory) {
    const targetInventory = target.inventory;

    if (targetInventory.packages.length > 0) {
      const stolenPackage = targetInventory.packages[0];
      inventory.money += stolenPackage.value;
      targetInventory.packages = targetInventory.packages.filter(
        (p) => p !== stolenPackage,
      );
      const target = game.gameData.entities.find(
        (t) => t.id === stolenPackage.targetId,
      );
      if (target) target.deletionFlag = true;
      return;
    }

    if (targetInventory.money > 0) {
      const stolenMoney = Math.min(targetInventory.money, elapsedTicks);
      game.gameData.stats.stolenEnemyMoney += stolenMoney;
      targetInventory.money -= stolenMoney;
      inventory.money += stolenMoney;
      return;
    }

    if (targetInventory.clothes !== "underwear") {
      if (ai.remainingUndressingTime <= 0) {
        ai.remainingUndressingTime = ai.undressingTime;
        return;
      }
      ai.remainingUndressingTime -= elapsedTicks;
      if (ai.remainingUndressingTime > 0) return;
      targetInventory.clothes = "underwear";
    }
  }

  ai.targetBeingCaptured = undefined;
  movement.enabled = true;
  pathing.enabled = true;
  return;
}

function checkTargetToChase(game: Game, entity: EntityType) {
  if (!entity.aiEnemy) return;
  if (!entity.mapPosition) return;
  if (!entity.mapPathing) return;
  if (!entity.mapMovement) return;
  if (!entity.characterSpriteRenderUpdater) return;
  const ai = entity.aiEnemy;
  const position = entity.mapPosition;
  const pathing = entity.mapPathing;
  const movement = entity.mapMovement;
  const renderer = entity.characterSpriteRenderUpdater;

  if (ai.targetBeingCaptured) return;

  if (entity.capturable && entity.capturable.state !== "free") return;

  const possibleTarget = getCharacterToChase(
    game,
    position,
    renderer.direction,
    ai.maxChaseDistance,
  );
  const target = possibleTarget?.character;
  if (!target || !target.mapPosition) {
    renderer.status = undefined;
    pathing.enabled = true;
    movement.enabled = true;
    movement.speed = ai.patrolMovementSpeed;
    return;
  }

  const targetPosition = target.mapPosition;
  if (possibleTarget.distance < ai.maxCaptureDistance) {
    if (!target.capturable) return;
    const direction = getDirectionFromTargetToTarget(position, targetPosition);
    renderer.direction = direction;
    if (target.characterSpriteRenderUpdater)
      target.characterSpriteRenderUpdater.direction = direction;
    if (target.mapMovement) target.mapMovement.enabled = false;
    target.capturable.state = "grabbed";
    ai.targetBeingCaptured = target.id;
    ai.remainingCaptureTime = ai.capturingTime;
    pathing.enabled = false;
    movement.enabled = false;
    return;
  }

  renderer.status = "warning";

  pathing.enabled = false;
  movement.enabled = true;
  movement.speed = ai.chaseMovementSpeed;

  const requestedPositions = [];
  if (targetPosition.y < position.y)
    requestedPositions.push({
      x: Math.round(position.x),
      y: Math.round(position.y) - 1,
    });
  if (targetPosition.y > position.y)
    requestedPositions.push({
      x: Math.round(position.x),
      y: Math.round(position.y) + 1,
    });
  if (targetPosition.x < position.x)
    requestedPositions.push({
      x: Math.round(position.x) - 1,
      y: Math.round(position.y),
    });
  if (targetPosition.x > position.x)
    requestedPositions.push({
      x: Math.round(position.x) + 1,
      y: Math.round(position.y),
    });

  const map = game.gameData.entities.find((e) => e.mapMetadata);
  if (!map) return;
  if (!map.mapMetadata) return;
  for (const tile of requestedPositions) {
    if (!map.mapMetadata.roads.some((r) => r.x === tile.x && r.y === tile.y))
      continue;
    movement.x = tile.x;
    movement.y = tile.y;
    return;
  }

  renderer.direction = getDirectionFromTargetToTarget(position, targetPosition);
}

function checkTargetCampArrival(game: Game, entity: EntityType) {
  if (!entity.aiEnemy) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPathing) return;
  const ai = entity.aiEnemy;
  const pathing = entity.mapPathing;
  const movement = entity.mapMovement;
  if (ai.targetBeingCaptured) return;
  if (!movement.enabled) return;
  if (!pathing.enabled) return;

  const startPosition = pathing.path[0];
  if (movement.x !== startPosition.x) {
    ai.leftForPatrol = true;
    return;
  }
  if (movement.y !== startPosition.y) {
    ai.leftForPatrol = true;
    return;
  }
  if (!ai.leftForPatrol) return;

  game.gameData.stats.gatheredEnemyMoney += entity.inventory?.money || 0;
  entity.deletionFlag = true;
}
