import { ControllerSystemMethod, Game } from "../engine";
import { EntityType } from "../entities";
import {
  distanceBetween,
  getDirectionFromTargetToTarget,
  getIsReachableMethod,
} from "../utils/analyzer";
import { getTilePathFromStartToFinish } from "../utils/pathing";

export function controllerAIGuardSystem(): ControllerSystemMethod {
  return (game, elapsedTicks) => {
    for (const entity of game.gameData.entities) {
      if (!entity.aiGuard) continue;

      checkRetargetToNearestTown(game, entity);
      checkStruggling(elapsedTicks, entity);
      if (entity.capturable?.state !== "free") {
        if (entity.mapPathing) entity.mapPathing.enabled = false;
        if (entity.mapMovement) entity.mapMovement.enabled = false;
        checkTargetTownArrival(game, entity);
        continue;
      }
      if (entity.inventory?.clothes === "underwear") {
        checkTargetTownArrival(game, entity);
        continue;
      }
      checkTargetBeingChased(game, elapsedTicks, entity);
      checkTargetBeingFreed(game, elapsedTicks, entity);
      checkTargetToChaseOrFree(game, entity);
      checkTargetTownArrival(game, entity);
    }
  };
}

function checkRetargetToNearestTown(game: Game, entity: EntityType) {
  if (!entity.aiGuard) return;
  if (!entity.inventory) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPosition) return;
  if (!entity.mapPathing) return;

  const ai = entity.aiGuard;
  const inventory = entity.inventory;
  const movement = entity.mapMovement;
  const pathing = entity.mapPathing;
  const position = entity.mapPosition;

  if (!pathing.enabled) return;

  if (ai.hasRetargetedNearestTown) {
    movement.speed = ai.escapeMovementSpeed;
    return;
  }

  if (inventory.clothes === "underwear") {
    const map = game.gameData.entities.find((e) => e.mapMetadata);
    if (!map) return;
    if (!map.mapMetadata) return;

    movement.speed = ai.escapeMovementSpeed;

    let targets = [];
    for (const entity of game.gameData.entities) {
      if (!entity.capturePrevention) continue;
      if (!entity.mapPosition) continue;
      const distance = distanceBetween(position, entity.mapPosition);
      targets.push({ entity, distance });
    }

    targets.sort((t1, t2) => t1.distance - t2.distance);

    for (const { entity } of targets) {
      if (!entity.mapPosition) continue;
      const target = {
        x: Math.round(entity.mapPosition.x - 0.8),
        y: Math.round(entity.mapPosition.y + 0.5),
      };
      const path = getTilePathFromStartToFinish(
        map.mapMetadata.roads,
        { x: Math.round(position.x), y: Math.round(position.y) },
        target,
      );
      if (path.length < 1) continue;
      const lastPath = path[path.length - 1];
      if (lastPath.x !== target.x || lastPath.y !== target.y) continue;

      pathing.path = path;
      pathing.step = 0;
      pathing.direction = "normal";
      movement.x = path[0].x;
      movement.y = path[0].y;
      ai.targetTown = entity.id;
      ai.hasRetargetedNearestTown = true;
      return;
    }
    return;
  }
}

function getCharacterToChase(
  game: Game,
  position: { x: number; y: number },
  direction: "n" | "s" | "e" | "w",
  maxChaseDistance: number,
): { character: EntityType; distance: number } | undefined {
  const reachable = getIsReachableMethod(
    position,
    direction,
    maxChaseDistance,
    1.5,
    0.5,
  );

  const actualTargets: EntityType[] = [];
  for (const target of game.gameData.entities) {
    if (!target.capturable) continue;
    if (!target.capturable.isCapturableByTownie) continue;
    if (target.capturable.state !== "free") continue;
    if (!target.mapPosition) continue;
    const position = target.mapPosition;
    if (direction === "n" && position.y < position.y) continue;
    if (direction === "s" && position.y > position.y) continue;
    if (direction === "w" && position.x < position.x) continue;
    if (direction === "e" && position.x > position.x) continue;
    actualTargets.push(target);
  }

  let closest: { character: EntityType; distance: number } | undefined;
  for (const target of actualTargets) {
    if (!target.mapPosition) continue;
    if (!reachable(target.mapPosition)) continue;
    const d = distanceBetween(target.mapPosition, position);
    if (d > maxChaseDistance) continue;
    if (closest && closest.distance < d) continue;
    closest = { character: target, distance: d };
  }
  return closest;
}

function getCharacterToFree(
  game: Game,
  position: { x: number; y: number },
  direction: "n" | "s" | "e" | "w",
  maxChaseDistance: number,
): { character: EntityType; distance: number } | undefined {
  const reachable = getIsReachableMethod(
    position,
    direction,
    maxChaseDistance,
    1.5,
    0.5,
  );

  const actualTargets: EntityType[] = [];
  for (const target of game.gameData.entities) {
    if (!target.capturable) continue;
    if (!target.capturable.isFreeableByTownie) continue;
    if (target.capturable.state !== "bound") continue;
    if (!target.mapPosition) continue;
    const position = target.mapPosition;
    if (direction === "n" && position.y < position.y) continue;
    if (direction === "s" && position.y > position.y) continue;
    if (direction === "w" && position.x < position.x) continue;
    if (direction === "e" && position.x > position.x) continue;
    actualTargets.push(target);
  }

  let closest: { character: EntityType; distance: number } | undefined;
  for (const target of actualTargets) {
    if (!target.mapPosition) continue;
    if (!reachable(target.mapPosition)) continue;
    const d = distanceBetween(target.mapPosition, position);
    if (d > maxChaseDistance) continue;
    if (closest && closest.distance < d) continue;
    closest = { character: target, distance: d };
  }
  return closest;
}

function checkStruggling(elapsedTicks: number, entity: EntityType) {
  if (!entity.aiGuard) return;
  if (!entity.capturable) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPathing) return;
  if (!entity.characterSpriteRenderUpdater) return;
  const ai = entity.aiGuard;
  const capturable = entity.capturable;
  const movement = entity.mapMovement;
  const pathing = entity.mapPathing;
  const renderer = entity.characterSpriteRenderUpdater;

  if (capturable.state !== "bound") {
    renderer.struggleDataPercent = undefined;
    return;
  }

  movement.enabled = false;
  pathing.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;
  renderer.status = "shouting";

  capturable.remainingTightness -= elapsedTicks * ai.struggleStrength;
  renderer.struggleDataPercent =
    capturable.remainingTightness / capturable.totalTightness;
  if (capturable.remainingTightness > 0) return;
  capturable.state = "free";
  movement.enabled = true;
  pathing.enabled = true;
}

function checkTargetBeingChased(
  game: Game,
  elapsedTicks: number,
  entity: EntityType,
) {
  if (!entity.aiGuard) return;
  if (!entity.inventory) return;
  if (!entity.mapPosition) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPathing) return;
  if (!entity.characterSpriteRenderUpdater) return;
  const ai = entity.aiGuard;
  const movement = entity.mapMovement;
  const pathing = entity.mapPathing;
  const inventory = entity.inventory;
  const renderer = entity.characterSpriteRenderUpdater;
  if (!ai.targetBeingChased) return;

  movement.enabled = false;
  pathing.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;

  const targetId = ai.targetBeingChased;
  const target = game.gameData.entities.find((e) => e.id === targetId);
  if (!target) {
    ai.targetBeingChased = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  if (!target.capturable) {
    ai.targetBeingChased = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  if (entity.capturable && entity.capturable.state !== "free") {
    if (target.capturable.state === "grabbed") target.capturable.state = "free";
    if (target.mapMovement) target.mapMovement.enabled = true;
    ai.targetBeingChased = undefined;
    movement.enabled = false;
    pathing.enabled = false;
    return;
  }

  if (target.capturable.state === "grabbed") {
    if (ai.remainingCaptureTime > 0) {
      ai.remainingCaptureTime -= elapsedTicks;
      return;
    }
    target.capturable.state = "bound";
    target.capturable.totalTightness = ai.captureStrength;
    target.capturable.remainingTightness = ai.captureStrength;
  }

  if (target.inventory) {
    const targetInventory = target.inventory;

    if (targetInventory.packages.length > 0) {
      const stolenPackage = targetInventory.packages[0];
      inventory.money += stolenPackage.value;
      targetInventory.packages = targetInventory.packages.filter(
        (p) => p !== stolenPackage,
      );
      const target = game.gameData.entities.find(
        (t) => t.id === stolenPackage.targetId,
      );
      if (target) target.deletionFlag = true;
      return;
    }

    if (targetInventory.money > 0) {
      const stolenMoney = Math.min(targetInventory.money, elapsedTicks);
      targetInventory.money -= stolenMoney;
      inventory.money += stolenMoney;
      return;
    }
  }

  ai.targetBeingChased = undefined;
  movement.enabled = true;
  pathing.enabled = true;
  return;
}

function checkTargetBeingFreed(
  game: Game,
  elapsedTicks: number,
  entity: EntityType,
) {
  if (!entity.aiGuard) return;
  if (!entity.inventory) return;
  if (!entity.mapPosition) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPathing) return;
  if (!entity.characterSpriteRenderUpdater) return;
  const ai = entity.aiGuard;
  const movement = entity.mapMovement;
  const pathing = entity.mapPathing;
  const renderer = entity.characterSpriteRenderUpdater;
  if (!ai.targetBeingFreed) return;

  if (entity.capturable && entity.capturable.state !== "free") {
    ai.targetBeingFreed = undefined;
    movement.enabled = false;
    pathing.enabled = false;
    return;
  }

  movement.enabled = false;
  pathing.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;

  const targetId = ai.targetBeingFreed;
  const target = game.gameData.entities.find((e) => e.id === targetId);
  if (!target) {
    ai.targetBeingFreed = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  if (!target.capturable) {
    ai.targetBeingFreed = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  if (target.capturable.state !== "bound") {
    ai.targetBeingFreed = undefined;
    movement.enabled = true;
    pathing.enabled = true;
    return;
  }

  target.capturable.remainingTightness -= elapsedTicks;
  if (target.capturable.remainingTightness > 0) return;
  ai.targetBeingFreed = undefined;
  target.capturable.state = "free";
  if (target.mapMovement) target.mapMovement.enabled = true;
  movement.enabled = true;
  pathing.enabled = true;
  return;
}

function checkTargetToChaseOrFree(game: Game, entity: EntityType) {
  if (!entity.aiGuard) return;
  if (!entity.mapPosition) return;
  if (!entity.mapPathing) return;
  if (!entity.mapMovement) return;
  if (!entity.characterSpriteRenderUpdater) return;
  const ai = entity.aiGuard;
  const position = entity.mapPosition;
  const pathing = entity.mapPathing;
  const movement = entity.mapMovement;
  const renderer = entity.characterSpriteRenderUpdater;

  if (ai.targetBeingChased) return;
  if (ai.targetBeingFreed) return;

  if (entity.capturable && entity.capturable.state !== "free") return;

  const possibleChaseTarget = getCharacterToChase(
    game,
    position,
    renderer.direction,
    ai.maxChaseDistance,
  );
  const targetToChase = possibleChaseTarget?.character;
  if (targetToChase && targetToChase.mapPosition) {
    const targetPosition = targetToChase.mapPosition;
    if (possibleChaseTarget.distance < ai.maxCaptureDistance) {
      if (!targetToChase.capturable) return;
      const direction = getDirectionFromTargetToTarget(
        position,
        targetPosition,
      );
      renderer.direction = direction;
      if (targetToChase.characterSpriteRenderUpdater)
        targetToChase.characterSpriteRenderUpdater.direction = direction;
      if (targetToChase.mapMovement) targetToChase.mapMovement.enabled = false;
      targetToChase.capturable.state = "grabbed";
      ai.targetBeingChased = targetToChase.id;
      ai.remainingCaptureTime = ai.capturingTime;
      pathing.enabled = false;
      movement.enabled = false;
      return;
    }

    renderer.status = "warning";

    pathing.enabled = false;
    movement.enabled = true;
    movement.speed = ai.chaseMovementSpeed;

    const requestedPositions = [];
    if (targetPosition.y < position.y)
      requestedPositions.push({
        x: Math.round(position.x),
        y: Math.round(position.y) - 1,
      });
    if (targetPosition.y > position.y)
      requestedPositions.push({
        x: Math.round(position.x),
        y: Math.round(position.y) + 1,
      });
    if (targetPosition.x < position.x)
      requestedPositions.push({
        x: Math.round(position.x) - 1,
        y: Math.round(position.y),
      });
    if (targetPosition.x > position.x)
      requestedPositions.push({
        x: Math.round(position.x) + 1,
        y: Math.round(position.y),
      });

    const map = game.gameData.entities.find((e) => e.mapMetadata);
    if (!map) return;
    if (!map.mapMetadata) return;
    for (const tile of requestedPositions) {
      if (!map.mapMetadata.roads.some((r) => r.x === tile.x && r.y === tile.y))
        continue;
      movement.x = tile.x;
      movement.y = tile.y;
      return;
    }

    renderer.direction = getDirectionFromTargetToTarget(
      position,
      targetPosition,
    );
    return;
  }

  const possibleFreeTarget = getCharacterToFree(
    game,
    position,
    renderer.direction,
    ai.maxChaseDistance,
  );
  const targetToFree = possibleFreeTarget?.character;
  if (targetToFree && targetToFree.mapPosition) {
    const targetPosition = targetToFree.mapPosition;
    if (possibleFreeTarget.distance < ai.maxFreeDistance) {
      if (!targetToFree.capturable) return;
      const direction = getDirectionFromTargetToTarget(
        position,
        targetPosition,
      );
      renderer.direction = direction;
      if (targetToFree.characterSpriteRenderUpdater)
        targetToFree.characterSpriteRenderUpdater.direction = direction;
      if (targetToFree.mapMovement) targetToFree.mapMovement.enabled = false;
      ai.targetBeingFreed = targetToFree.id;
      pathing.enabled = false;
      movement.enabled = false;
      return;
    }

    renderer.status = "warning";

    pathing.enabled = false;
    movement.enabled = true;
    movement.speed = ai.chaseMovementSpeed;

    const requestedPositions = [];
    if (targetPosition.y < position.y)
      requestedPositions.push({
        x: Math.round(position.x),
        y: Math.round(position.y) - 1,
      });
    if (targetPosition.y > position.y)
      requestedPositions.push({
        x: Math.round(position.x),
        y: Math.round(position.y) + 1,
      });
    if (targetPosition.x < position.x)
      requestedPositions.push({
        x: Math.round(position.x) - 1,
        y: Math.round(position.y),
      });
    if (targetPosition.x > position.x)
      requestedPositions.push({
        x: Math.round(position.x) + 1,
        y: Math.round(position.y),
      });

    const map = game.gameData.entities.find((e) => e.mapMetadata);
    if (!map) return;
    if (!map.mapMetadata) return;
    for (const tile of requestedPositions) {
      if (!map.mapMetadata.roads.some((r) => r.x === tile.x && r.y === tile.y))
        continue;
      movement.x = tile.x;
      movement.y = tile.y;
      return;
    }

    renderer.direction = getDirectionFromTargetToTarget(
      position,
      targetPosition,
    );
    return;
  }
  renderer.status = undefined;
  pathing.enabled = true;
  movement.enabled = true;
  movement.speed = ai.patrolMovementSpeed;
}

function checkTargetTownArrival(game: Game, entity: EntityType) {
  if (!entity.aiGuard) return;
  if (!entity.mapPosition) return;
  const ai = entity.aiGuard;
  const position = entity.mapPosition;

  const targetTownId = ai.targetTown;
  const targetTown = game.gameData.entities.find((e) => e.id === targetTownId);
  if (!targetTown) return;
  if (!targetTown.mapPosition) return;
  if (Math.round(targetTown.mapPosition.y) !== Math.round(position.y)) return;
  if (targetTown.mapPosition.x < position.x - 1) return;
  if (targetTown.mapPosition.x > position.x + 1) return;

  game.gameData.stats.gatheredTownieMoney += entity.inventory?.money || 0;
  entity.deletionFlag = true;
}
