import { ControllerSystemMethod, Game } from "../engine";
import {
  distanceBetween,
  getDirectionFromTargetToTarget,
  getIsReachableMethod,
} from "../utils/analyzer";
import { EntityType } from "../entities";
import { createDeliveryTargetEntity } from "../entities/delivery-target";

export function controllerControlsPlayerSystem(): ControllerSystemMethod {
  return (game, elapsedTicks) => {
    for (const entity of game.gameData.entities) {
      if (!entity.controlsPlayer) continue;
      if (!entity.characterSpriteRenderUpdater) continue;
      entity.characterSpriteRenderUpdater.textData = undefined;

      updateActionKeys(game, entity);

      updateCanReequip(game, entity);
      updateReequipping(elapsedTicks, entity);

      updateCanCapture(game, elapsedTicks, entity);
      updateCapturing(game, elapsedTicks, entity);

      updateCanFree(game, entity);
      updateFreeing(game, elapsedTicks, entity);

      updateCanRob(game, entity);
      updateRobbing(game, elapsedTicks, entity);

      updateCanInteractWithNPC(game, entity);
      updateInteractWithNPC(game, elapsedTicks, entity);

      updateStruggling(game, entity);

      updateRequestedPosition(game, entity);
    }
  };
}

function updateActionKeys(game: Game, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  const controls = entity.controlsPlayer;
  if (controls.holdingAction1Key) {
    controls.pressedAction1KeyThisFrame = false;
    controls.holdingAction1Key = game.controls.keys.action1;
  } else {
    controls.pressedAction1KeyThisFrame = game.controls.keys.action1;
    controls.holdingAction1Key = game.controls.keys.action1;
  }
}

function updateCanReequip(game: Game, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.inventory) return;
  if (!entity.mapPosition) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const inventory = entity.inventory;
  const controls = entity.controlsPlayer;
  const position = entity.mapPosition;
  const renderer = entity.characterSpriteRenderUpdater;
  if (capture.state !== "free") return;
  if (controls.captureTarget) return;
  if (controls.equip.current > 0) return;
  if (inventory.clothes !== "underwear") return;

  const targetResupply = game.gameData.entities.find(
    (e) =>
      e.capturePrevention &&
      e.capturePrevention.enabled &&
      e.mapPosition &&
      distanceBetween(e.mapPosition, position) < e.capturePrevention.radius,
  );
  if (!targetResupply) return;

  renderer.textData = "gearUp";

  if (controls.pressedAction1KeyThisFrame) {
    controls.equip.current = controls.equip.time;
  }
}

function updateReequipping(elapsedTicks: number, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.inventory) return;
  if (!entity.mapMovement) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const inventory = entity.inventory;
  const controls = entity.controlsPlayer;
  const movement = entity.mapMovement;
  const renderer = entity.characterSpriteRenderUpdater;

  if (controls.equip.current <= 0) return;

  movement.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 4;
  controls.equip.current -= elapsedTicks;
  renderer.direction = (["s", "w", "n", "e"] as const)[
    Math.max(
      0,
      Math.floor(controls.equip.current * controls.equipRotationSpeed) % 4,
    )
  ];

  if (controls.equip.current <= 0) {
    movement.enabled = true;
    inventory.clothes = "dress";
  }
}

function updateCanCapture(
  game: Game,
  elapsedTicks: number,
  entity: EntityType,
) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.inventory) return;
  if (!entity.mapPosition) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const inventory = entity.inventory;
  const controls = entity.controlsPlayer;
  const position = entity.mapPosition;
  const renderer = entity.characterSpriteRenderUpdater;

  if (capture.state !== "free") return;
  if (inventory.clothes === "underwear") return;
  if (controls.preventCapturingOthers) return;
  if (controls.captureTarget) return;
  if (controls.equip.current > 0) return;

  if (controls.captureAttempt.current > 0) {
    controls.captureAttempt.current -= elapsedTicks;
    if (controls.captureAttempt.current < 0)
      controls.captureAttempt.current = 0;
    return;
  }

  if (controls.isBandit) {
    const capturePreventions = game.gameData.entities
      .filter((e) => e.mapPosition && e.capturePrevention)
      .map((e) => ({
        x: e.mapPosition!.x,
        y: e.mapPosition!.y,
        radius: e.capturePrevention!.radius,
      }));
    const isNearCaptureBlocker = capturePreventions.some(
      (t) => distanceBetween(position, t) < t.radius,
    );
    if (isNearCaptureBlocker) return;
  }

  const playerCanReach = getIsReachableMethod(
    position,
    renderer.direction,
    controls.captureDistanceFront,
    controls.captureDistanceSides,
    0.5,
  );
  const targetEntity = game.gameData.entities.find(
    (e) =>
      e.capturable &&
      ((!controls.isBandit && e.capturable.isCapturableByTownie) ||
        (controls.isBandit && e.capturable.isCapturableByBandits)) &&
      e.capturable.state === "free" &&
      e.mapPosition &&
      playerCanReach(e.mapPosition),
  );
  if (!targetEntity) {
    if (controls.pressedAction1KeyThisFrame) {
      controls.captureAttempt.current = controls.captureAttempt.time;
    }
    return;
  }

  renderer.textData = "capture";

  if (controls.pressedAction1KeyThisFrame && targetEntity.capturable) {
    if (!targetEntity.mapPosition) return;
    if (!targetEntity.capturable) return;
    if (targetEntity.characterSpriteRenderUpdater) {
      const direction = getDirectionFromTargetToTarget(
        position,
        targetEntity.mapPosition,
      );
      const targetRenderer = targetEntity.characterSpriteRenderUpdater;
      targetRenderer.direction = direction;
      renderer.direction = direction;
    }
    targetEntity.capturable.state = "grabbed";
    controls.captureTarget = targetEntity.id;
    controls.capture.current = controls.capture.time;
  }
}

function updateCapturing(game: Game, elapsedTicks: number, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapMovement) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const movement = entity.mapMovement;
  const renderer = entity.characterSpriteRenderUpdater;

  if (!controls.captureTarget) return;

  const targetEntity = game.gameData.entities.find(
    (e) => e.id === controls.captureTarget,
  );
  if (!targetEntity) return;
  if (!targetEntity.capturable) return;
  if (capture.state !== "free" || controls.preventCapturingOthers) {
    controls.captureTarget = undefined;
    targetEntity.capturable.state = "free";
    return;
  }

  movement.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;

  controls.capture.current -= elapsedTicks;

  if (controls.capture.current <= 0) {
    controls.captureTarget = undefined;
    targetEntity.capturable.state = "bound";
    targetEntity.capturable.remainingTightness = controls.captureStrength;
    targetEntity.capturable.totalTightness = controls.captureStrength;
    movement.enabled = true;
  }
}

function updateCanFree(game: Game, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapPosition) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const position = entity.mapPosition;
  const renderer = entity.characterSpriteRenderUpdater;

  if (controls.captureTarget) return;
  if (controls.equip.current > 0) return;
  if (capture.state !== "free") return;

  const playerCanReach = getIsReachableMethod(
    position,
    renderer.direction,
    controls.freeDistanceFront,
    controls.freeDistanceSides,
    0.5,
  );
  const targetEntity = game.gameData.entities.find(
    (e) =>
      e.capturable &&
      ((!controls.isBandit && e.capturable.isFreeableByTownie) ||
        (controls.isBandit && e.capturable.isFreeableByBandits)) &&
      e.capturable.state === "bound" &&
      e.mapPosition &&
      playerCanReach(e.mapPosition),
  );
  if (!targetEntity) return;

  renderer.textData = "untie";

  if (controls.pressedAction1KeyThisFrame && targetEntity.capturable) {
    controls.freeTarget = targetEntity.id;
  }
}

function updateFreeing(game: Game, elapsedTicks: number, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapMovement) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const movement = entity.mapMovement;
  const renderer = entity.characterSpriteRenderUpdater;

  if (!controls.freeTarget) return;

  const targetEntity = game.gameData.entities.find(
    (e) => e.id === controls.freeTarget,
  );
  if (!targetEntity) return;
  if (!targetEntity.capturable) return;
  if (!targetEntity.mapMovement) return;
  if (capture.state !== "free" || controls.preventCapturingOthers) {
    controls.freeTarget = undefined;
    return;
  }

  movement.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;

  targetEntity.capturable.remainingTightness -= elapsedTicks;

  if (targetEntity.capturable.remainingTightness <= 0) {
    controls.freeTarget = undefined;
    targetEntity.capturable.state = "free";
    targetEntity.mapMovement.enabled = true;
    movement.enabled = true;
  }
}

function updateCanRob(game: Game, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapPosition) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const position = entity.mapPosition;
  const renderer = entity.characterSpriteRenderUpdater;

  if (!controls.isBandit) return;
  if (controls.captureTarget) return;
  if (controls.equip.current > 0) return;
  if (capture.state !== "free") return;

  const playerCanReach = getIsReachableMethod(
    position,
    renderer.direction,
    controls.robDistanceFront,
    controls.robDistanceSides,
    0.5,
  );
  const targetEntity = game.gameData.entities.find(
    (e) =>
      e.capturable &&
      e.capturable.isCapturableByBandits &&
      e.capturable.state === "bound" &&
      e.inventory &&
      (e.inventory.money > 0 || e.inventory.clothes !== "underwear") &&
      e.mapPosition &&
      playerCanReach(e.mapPosition),
  );
  if (!targetEntity) return;

  renderer.textData = "rob";

  if (controls.pressedAction1KeyThisFrame && targetEntity.capturable) {
    controls.robTarget = targetEntity.id;
  }
}

function updateRobbing(game: Game, elapsedTicks: number, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapMovement) return;
  if (!entity.characterSpriteRenderUpdater) return;
  if (!entity.inventory) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const movement = entity.mapMovement;
  const renderer = entity.characterSpriteRenderUpdater;
  const inventory = entity.inventory;

  if (!controls.isBandit) return;
  if (!controls.robTarget) return;

  const targetEntity = game.gameData.entities.find(
    (e) => e.id === controls.robTarget,
  );
  if (!targetEntity) return;
  if (!targetEntity.capturable) return;
  if (!targetEntity.mapMovement) return;
  if (!targetEntity.inventory) return;
  if (capture.state !== "free") {
    controls.robTarget = undefined;
    return;
  }

  movement.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;

  const targetInventory = targetEntity.inventory;

  if (targetInventory.packages.length > 0) {
    const stolenPackage = targetInventory.packages[0];
    inventory.money += stolenPackage.value;
    targetInventory.packages = targetInventory.packages.filter(
      (p) => p !== stolenPackage,
    );
    const target = game.gameData.entities.find(
      (t) => t.id === stolenPackage.targetId,
    );
    if (target) target.deletionFlag = true;
    return;
  }

  if (targetInventory.money > 0) {
    const stolenMoney = Math.min(targetInventory.money, elapsedTicks);
    game.gameData.stats.gatheredPlayerMoney += stolenMoney;
    targetInventory.money -= stolenMoney;
    inventory.money += stolenMoney;
    return;
  }

  if (targetInventory.clothes !== "underwear") {
    if (controls.remainingUndressingTime <= 0) {
      controls.remainingUndressingTime = controls.undressingTime;
      return;
    }
    controls.remainingUndressingTime -= elapsedTicks;
    if (controls.remainingUndressingTime > 0) return;
    targetInventory.clothes = "underwear";
  }

  controls.robTarget = undefined;
  movement.enabled = true;
}

function updateCanInteractWithNPC(game: Game, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapPosition) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const position = entity.mapPosition;
  const renderer = entity.characterSpriteRenderUpdater;

  if (controls.captureTarget) return;
  if (controls.equip.current > 0) return;
  if (capture.state !== "free") return;

  const canReach = getIsReachableMethod(
    position,
    renderer.direction,
    controls.npcInteractionDistanceFront,
    controls.npcInteractionDistanceSides,
    0.5,
  );
  const npc = game.gameData.entities.find(
    (e) => e.interactibleNPC && e.mapPosition && canReach(e.mapPosition),
  );
  if (!npc) return;
  if (!npc.interactibleNPC) return;

  if (npc.interactibleNPC.type === "pickUp")
    renderer.textData = "pickUpDelivery";
  if (npc.interactibleNPC.type === "deliver") renderer.textData = "deliver";

  if (controls.pressedAction1KeyThisFrame) {
    controls.npcInteractionTarget = npc.id;
    controls.npcInteraction.current = controls.npcInteraction.time;
    if (npc.mapMovement) npc.mapMovement.enabled = false;
  }
}

function updateInteractWithNPC(
  game: Game,
  elapsedTicks: number,
  entity: EntityType,
) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.inventory) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPosition) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const inventory = entity.inventory;
  const movement = entity.mapMovement;
  const position = entity.mapPosition;
  const renderer = entity.characterSpriteRenderUpdater;

  if (!controls.npcInteractionTarget) return;

  const npc = game.gameData.entities.find(
    (e) => e.id === controls.npcInteractionTarget,
  );
  if (!npc || !npc.interactibleNPC) {
    renderer.status = undefined;
    controls.npcInteractionTarget = undefined;
    movement.enabled = true;
    return;
  }

  if (
    controls.captureTarget ||
    controls.equip.current > 0 ||
    capture.state !== "free"
  ) {
    renderer.status = undefined;
    controls.npcInteractionTarget = undefined;
    if (npc.mapMovement) npc.mapMovement.enabled = true;
    movement.enabled = true;
    return;
  }

  movement.enabled = false;
  renderer.movement.isAnimated = true;
  renderer.movement.animationSpeed = 6;
  renderer.status = "warning";
  if (npc.mapPosition) {
    renderer.direction = getDirectionFromTargetToTarget(
      position,
      npc.mapPosition,
    );
  }

  const targetRenderer = npc.characterSpriteRenderUpdater;
  if (targetRenderer) {
    targetRenderer.movement.isAnimated = true;
    targetRenderer.movement.animationSpeed = 6;
    targetRenderer.status = "warning";
    if (npc.mapPosition) {
      targetRenderer.direction = getDirectionFromTargetToTarget(
        npc.mapPosition,
        position,
      );
    }
  }

  controls.npcInteraction.current -= elapsedTicks;

  if (controls.npcInteraction.current <= 0) {
    renderer.status = undefined;
    controls.npcInteractionTarget = undefined;
    if (npc.mapMovement) npc.mapMovement.enabled = true;
    movement.enabled = true;

    if (npc.interactibleNPC.type === "pickUp") {
      if (!npc.inventory) return;
      for (const deliveryPackage of npc.inventory.packages) {
        const town = game.gameData.entities.find(
          (e) => e.id === deliveryPackage.targetId,
        );
        if (!town) continue;
        if (!town.mapPosition) continue;
        const map = game.gameData.entities.find((e) => e.mapMetadata);
        if (!map) continue;
        if (!map.mapMetadata) continue;
        let target = game.gameData.entities.find(
          (e) =>
            e.interactibleNPC &&
            e.interactibleNPC.type === "deliver" &&
            e.interactibleNPC.town === town.id,
        );
        if (!target) {
          target = createDeliveryTargetEntity(
            town.mapPosition.x + 0.2,
            town.mapPosition.y + 0.5,
            town.id,
            map.mapMetadata.tileSize,
          );
          game.gameData.entities.push(target);
        }

        inventory.packages.push({
          name: deliveryPackage.name,
          value: deliveryPackage.value,
          targetId: target.id,
        });
      }
      npc.deletionFlag = true;
    }
    if (npc.interactibleNPC.type === "deliver") {
      const matchingPackages = inventory.packages.filter(
        (p) => p.targetId === npc.id,
      );
      if (matchingPackages.length <= 0) return;
      for (const deliveryPackage of matchingPackages) {
        inventory.money += deliveryPackage.value;
        game.gameData.stats.gatheredPlayerMoney += deliveryPackage.value;
      }
      inventory.packages = inventory.packages.filter(
        (p) => p.targetId !== npc.id,
      );
      npc.deletionFlag = true;
    }
  }
}

function updateStruggling(game: Game, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapMovement) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const movement = entity.mapMovement;
  const renderer = entity.characterSpriteRenderUpdater;

  if (capture.state !== "free") {
    capture.isStruggling =
      game.controls.keys.up ||
      game.controls.keys.down ||
      game.controls.keys.left ||
      game.controls.keys.right;
  }
  if (capture.state === "bound") {
    const keys = ["up", "down", "left", "right"] as const;
    for (const key of keys) {
      if (game.controls.keys[key] && !controls.lastStruggleKey[key]) {
        capture.remainingTightness--;
      }
      controls.lastStruggleKey[key] = game.controls.keys[key];
    }

    renderer.struggleDataPercent =
      capture.remainingTightness / capture.totalTightness;
    renderer.textData = "struggle";
    if (capture.remainingTightness < 0) {
      capture.state = "free";
      capture.remainingTightness = 0;
      renderer.struggleDataPercent = undefined;
      renderer.status = undefined;
      movement.enabled = true;
    }
  } else {
    renderer.struggleDataPercent = undefined;
  }
}

function updateRequestedPosition(game: Game, entity: EntityType) {
  if (!entity.controlsPlayer) return;
  if (!entity.capturable) return;
  if (!entity.mapMovement) return;
  if (!entity.mapPosition) return;
  if (!entity.characterSpriteRenderUpdater) return;

  const capture = entity.capturable;
  const controls = entity.controlsPlayer;
  const movement = entity.mapMovement;
  const position = entity.mapPosition;
  const renderer = entity.characterSpriteRenderUpdater;

  if (capture.state !== "free") return;
  if (controls.captureTarget) return;
  if (controls.freeTarget) return;
  if (controls.equip.current > 0) return;

  if (game.controls.keys.action2) {
    movement.speed = controls.sprintSpeed;
  } else {
    movement.speed = controls.walkSpeed;
  }

  if (!movement.isOnPosition) return;

  const requestedPositions = [];
  if (game.controls.keys.up) {
    requestedPositions.push({
      x: Math.round(position.x),
      y: Math.round(position.y) - 1,
    });
  }
  if (game.controls.keys.down) {
    requestedPositions.push({
      x: Math.round(position.x),
      y: Math.round(position.y) + 1,
    });
  }
  if (game.controls.keys.left) {
    requestedPositions.push({
      x: Math.round(position.x) - 1,
      y: Math.round(position.y),
    });
  }
  if (game.controls.keys.right) {
    requestedPositions.push({
      x: Math.round(position.x) + 1,
      y: Math.round(position.y),
    });
  }

  const map = game.gameData.entities.find((e) => e.mapMetadata);
  if (!map) return;
  if (!map.mapMetadata) return;
  for (const tile of requestedPositions) {
    if (
      controls.preventMovingOffRoad &&
      !map.mapMetadata.roads.some((r) => r.x === tile.x && r.y === tile.y)
    )
      continue;
    if (
      controls.preventWalkingOverRestrainedAllies &&
      game.gameData.entities.some(
        (e) =>
          e.capturable &&
          e.mapPosition &&
          ((!controls.isBandit && e.capturable.isFreeableByTownie) ||
            (controls.isBandit && e.capturable.isFreeableByBandits)) &&
          e.capturable.state === "bound" &&
          Math.round(e.mapPosition.x) === tile.x &&
          Math.round(e.mapPosition.y) === tile.y,
      )
    )
      continue;
    movement.x = tile.x;
    movement.y = tile.y;
    return;
  }

  if (game.controls.keys.up) renderer.direction = "n";
  if (game.controls.keys.down) renderer.direction = "s";
  if (game.controls.keys.right) renderer.direction = "e";
  if (game.controls.keys.left) renderer.direction = "w";
}
