import { ControllerSystemMethod, Game } from "../engine";
import { EntityType } from "../entities";
import { createEnemyEntity } from "../entities/enemy";
import { distanceBetween } from "../utils/analyzer";
import { getTilePathFromStartToFinish } from "../utils/pathing";
import {
  getProbabilityResult,
  getRandomArrayItem,
  getRandomIntegerInRange,
} from "../utils/random";

export function controllerSpawnerEnemySystem(): ControllerSystemMethod {
  return (game, elapsedTicks) => {
    for (const entity of game.gameData.entities) {
      if (!entity.spawnerEnemy) continue;
      entity.spawnerEnemy.remainingTime -= elapsedTicks;
      if (entity.spawnerEnemy.remainingTime < 0) {
        if (getProbabilityResult(entity.spawnerEnemy.probability)) {
          spawnNewEnemy(game, entity, entity.spawnerEnemy.distanceFromPosition);
        }
        entity.spawnerEnemy.remainingTime = getRandomIntegerInRange(
          entity.spawnerEnemy.minTime,
          entity.spawnerEnemy.maxTime,
        );
      }
    }
  };
}

function getFurthestPointFrom(
  origin: { x: number; y: number },
  points: Array<{ x: number; y: number }>,
  maxDistance: number,
): { x: number; y: number; distance: number } | undefined {
  let foundPoint: { x: number; y: number; distance: number } | undefined;
  for (const point of points) {
    const d = distanceBetween(origin, point);
    if (d > maxDistance) continue;
    if (foundPoint && foundPoint.distance > d) continue;
    foundPoint = { x: point.x, y: point.y, distance: d };
  }
  return foundPoint;
}

function spawnNewEnemy(game: Game, camp: EntityType, maxDistance: number) {
  if (!camp.mapPosition) return;
  const map = game.gameData.entities.find((e) => e.mapMetadata);
  if (!map) return;
  if (!map.mapMetadata) return;
  const center = { x: camp.mapPosition.x, y: camp.mapPosition.y };
  const positions = [
    getFurthestPointFrom(
      center,
      map.mapMetadata.roads.filter((r) => r.x < center.x && r.y < center.y),
      maxDistance,
    ),
    getFurthestPointFrom(
      center,
      map.mapMetadata.roads.filter((r) => r.x >= center.x && r.y < center.y),
      maxDistance,
    ),
    getFurthestPointFrom(
      center,
      map.mapMetadata.roads.filter((r) => r.x < center.x && r.y >= center.y),
      maxDistance,
    ),
    getFurthestPointFrom(
      center,
      map.mapMetadata.roads.filter((r) => r.x >= center.x && r.y >= center.y),
      maxDistance,
    ),
  ];
  const validPositions = positions
    .filter((p): p is { x: number; y: number; distance: number } => !!p)
    .filter((p) => p.distance > 3);
  const target = getRandomArrayItem(validPositions);
  if (!target) return;
  const path = getTilePathFromStartToFinish(
    map.mapMetadata.roads,
    {
      x: Math.round(camp.mapPosition.x - 0.8),
      y: Math.round(camp.mapPosition.y + 0.5),
    },
    target,
  );
  if (path.length < 2) return;

  game.gameData.entities.push(
    createEnemyEntity(camp.id, path, map.mapMetadata.tileSize),
  );
}
