import { CharacterSpriteRenderUpdaterComponent } from "../components/character-sprite-render-updater";
import { Game, SystemMethod } from "../engine";
import { CharacterRenderingData } from "../images";
import { PIXI } from "../pixi";

export function rendererCharacterSpriteUpdaterSystem(): SystemMethod {
  return (game) => {
    for (const entity of game.gameData.entities) {
      if (!entity.characterSpriteRenderUpdater) continue;

      if (
        spriteDataAreIdentical(
          entity.characterSpriteRenderUpdater,
          entity.characterSpriteRenderUpdater.drawnSprite,
        )
      ) {
        continue;
      }

      entity.characterSpriteRenderUpdater.drawnSprite = {
        ...entity.characterSpriteRenderUpdater,
        movement: entity.characterSpriteRenderUpdater.movement.frame,
        expression: entity.characterSpriteRenderUpdater.expression.frame,
      };

      updatePixiContainer(
        game,
        entity.characterSpriteRenderUpdater,
        game.images.characterLarge,
      );
    }
  };
}

function spriteDataAreIdentical(
  sprite: CharacterSpriteRenderUpdaterComponent,
  lastSprite: CharacterSpriteRenderUpdaterComponent["drawnSprite"],
) {
  if (!lastSprite) return false;

  // Note: These are all "!=" value comparisons, as these are checking for output similarity and not identity similarity
  if (sprite.clothing != lastSprite.clothing) return false;
  if (sprite.direction != lastSprite.direction) return false;
  if (sprite.clothesColor != lastSprite.clothesColor) return false;
  if (sprite.expression.frame != lastSprite.expression) return false; // Only the frame is important here
  if (sprite.eyesColor != lastSprite.eyesColor) return false;
  if (sprite.hairColor != lastSprite.hairColor) return false;
  if (sprite.movement.frame != lastSprite.movement) return false; // Only the frame is important here
  if (sprite.selected != lastSprite.selected) return false;
  if (sprite.state != lastSprite.state) return false;
  if (sprite.status != lastSprite.status) return false;
  if (sprite.textData != lastSprite.textData) return false;
  if (sprite.underwearColor != lastSprite.underwearColor) return false;
  if (sprite.displayedMoney != lastSprite.displayedMoney) return false;
  if (sprite.struggleDataPercent != lastSprite.struggleDataPercent)
    return false;
  return true;
}

function getActualFrameIndexFrom(frame: 0 | 1 | 2 | 3): number {
  if (frame === 0) return 2;
  if (frame === 1) return 3;
  if (frame === 2) return 2;
  if (frame === 3) return 1;
  return 2;
}

function updatePixiContainer(
  game: Game,
  renderer: CharacterSpriteRenderUpdaterComponent,
  images: CharacterRenderingData,
) {
  const container = renderer.container;
  const movementFrameIndex = getActualFrameIndexFrom(renderer.movement.frame);
  const frame = (() => {
    let start = "struggle";
    if (renderer.direction === "n") start = "walkN";
    if (renderer.direction === "e") start = "walkE";
    if (renderer.direction === "s") start = "walkS";
    if (renderer.direction === "w") start = "walkW";
    if (renderer.state === "struggle") start = "struggle";
    start += movementFrameIndex;
    return start;
  })();
  container.removeChildren();
  container.addChild(new PIXI.Sprite(images.body.textures[frame]));

  const eyes = new PIXI.Sprite(images.eyes.textures[frame]);
  eyes.tint = renderer.eyesColor;
  container.addChild(eyes);

  if (renderer.clothing === "dress") {
    const dress = new PIXI.Sprite(images.dress.textures[frame]);
    dress.tint = renderer.clothesColor;
    container.addChild(dress);
  } else if (renderer.clothing === "armor") {
    const armor = new PIXI.Sprite(images.armor.textures[frame]);
    armor.tint = renderer.clothesColor;
    container.addChild(armor);
  } else if (renderer.clothing === "underwear") {
    const underwear = new PIXI.Sprite(images.underwear.textures[frame]);
    underwear.tint = renderer.underwearColor;
    container.addChild(underwear);
  }

  const hair = new PIXI.Sprite(images.hair.textures[frame]);
  hair.tint = renderer.hairColor;
  container.addChild(hair);

  container.addChild(new PIXI.Sprite(images.binds.textures[frame]));

  const expressionFrameIndex = getActualFrameIndexFrom(
    renderer.expression.frame,
  );
  if (renderer.selected) {
    let frame = "select";
    frame += expressionFrameIndex;
    container.addChild(new PIXI.Sprite(images.expressions.textures[frame]));
  }
  if (renderer.status === "shouting") {
    let frame = "shout";
    frame += expressionFrameIndex;
    container.addChild(new PIXI.Sprite(images.expressions.textures[frame]));
  }
  if (renderer.status === "warning") {
    let frame = "warning";
    frame += expressionFrameIndex;
    container.addChild(new PIXI.Sprite(images.expressions.textures[frame]));
  }

  if (renderer.textData === "gearUp") {
    const text = new PIXI.Text({
      text: "[SPACE] Gear up",
      style: {
        fontSize: 16,
        fill: "white",
      },
    });
    text.anchor.set(0.5, 0);
    text.x = eyes.width / 2;
    text.y -= 10;
    container.addChild(text);
  }
  if (renderer.textData === "untie") {
    const text = new PIXI.Text({
      text: "[SPACE] Untie",
      style: {
        fontSize: 16,
        fill: "white",
      },
    });
    text.anchor.set(0.5, 0);
    text.x = eyes.width / 2;
    text.y -= 10;
    container.addChild(text);
  }
  if (renderer.textData === "rob") {
    const text = new PIXI.Text({
      text: "[SPACE] Rob",
      style: {
        fontSize: 16,
        fill: "white",
      },
    });
    text.anchor.set(0.5, 0);
    text.x = eyes.width / 2;
    text.y -= 10;
    container.addChild(text);
  }
  if (renderer.textData === "capture") {
    const text = new PIXI.Text({
      text: "[SPACE] Capture",
      style: {
        fontSize: 16,
        fill: "white",
      },
    });
    text.anchor.set(0.5, 0);
    text.x = eyes.width / 2;
    text.y -= 10;
    container.addChild(text);
  }
  if (renderer.textData === "struggle") {
    const text = new PIXI.Text({
      text: "[ARROWS] Struggle out!",
      style: {
        fontSize: 16,
        fill: "white",
      },
    });
    text.anchor.set(0.5, 0);
    text.x = eyes.width / 2;
    text.y -= 10;
    container.addChild(text);
  }
  if (renderer.textData === "deliver") {
    const text = new PIXI.Text({
      text: "[SPACE] Deliver package",
      style: {
        fontSize: 16,
        fill: "white",
      },
    });
    text.anchor.set(0.5, 0);
    text.x = eyes.width / 2;
    text.y -= 10;
    container.addChild(text);
  }
  if (renderer.textData === "pickUpDelivery") {
    const text = new PIXI.Text({
      text: "[SPACE] Pick up package",
      style: {
        fontSize: 16,
        fill: "white",
      },
    });
    text.anchor.set(0.5, 0);
    text.x = eyes.width / 2;
    text.y -= 10;
    container.addChild(text);
  }
  if (renderer.struggleDataPercent !== undefined) {
    const captureContainer = new PIXI.Container();
    const captureContainerPercentBarBorder = new PIXI.Graphics();
    captureContainerPercentBarBorder.x = 0;
    captureContainerPercentBarBorder.y = -10;
    captureContainerPercentBarBorder.rect(2, 2, 86, 12);
    captureContainerPercentBarBorder.fill(0xffffff);
    captureContainerPercentBarBorder.stroke({ width: 2, color: 0xcccccc });
    captureContainer.addChild(captureContainerPercentBarBorder);
    const captureContainerPercentBar = new PIXI.Graphics();
    captureContainerPercentBar.x = 0;
    captureContainerPercentBar.y = -10;
    const length = 82 * renderer.struggleDataPercent;
    captureContainerPercentBar.rect(4, 4, length, 8);
    captureContainerPercentBar.fill(0xcc00aa);
    captureContainer.addChild(captureContainerPercentBar);
    container.addChild(captureContainer);
  }

  if (renderer.displayedMoney !== undefined) {
    const moneyContainer = new PIXI.Container();
    const moneySprite = new PIXI.Sprite(game.images.goldCoin);
    moneySprite.x = 15;
    moneySprite.y = -35;
    moneyContainer.addChild(moneySprite);
    const money = new PIXI.Text({
      text: Math.floor(renderer.displayedMoney),
      style: {
        fill: "white",
        fontSize: 18,
      },
    });
    money.x = 42;
    money.y = -30;
    moneyContainer.addChild(money);
    container.addChild(moneyContainer);
  }
}
