export function getRandom(): number {
  return Math.random();
}

export function getProbabilityResult(successProbability: number): boolean {
  // Sanity check with random chance of 75%
  // - Result 20% => 20% <= 75% => OK
  // - Result 40% => 40% <= 75% => OK
  // - Result 60% => 60% <= 75% => OK
  // - Result 80% => 80% <= 75% => Failed
  return getRandom() <= successProbability;
}

/**
 * @param {number} start The starting value (included in the result)
 * @param {number} end The end value (not included in the result)
 * @returns {number} An integer between start and end, inclusive
 */
export function getRandomIntegerInRange(start: number, end: number): number {
  return Math.floor(start + getRandom() * (end - start));
}

export function getRandomArrayItem<T>(array: readonly T[]): T {
  return array[getRandomIntegerInRange(0, array.length)];
}

/**
 * @param array The array to get data from
 * @param quantity The quantity of items to get from the array
 * @returns The array to get data out. If `array` has less elements than `quantity`, will match `array` exactly.
 */
export function getSeveralRandomArrayItems<T>(
  array: readonly T[],
  quantity: number,
): T[] {
  if (quantity <= 0) return [];
  if (array.length < quantity) return [...array];
  let tempArray = [...array];
  const result: T[] = [];
  for (let i = 0; i < quantity; ++i) {
    const index = getRandomIntegerInRange(0, tempArray.length);
    result.push(tempArray[index]);
    tempArray = tempArray.filter((_, fi) => fi !== index);
  }
  return result;
}

export type WeightedArray<T> = Array<{ weight: number; item: T }>;
export type ReadonlyWeightedArray<T> = ReadonlyArray<WeightedArray<T>[number]>;

export function getRandomItemInWeightedArray<T>(
  array: ReadonlyWeightedArray<T>,
): T | undefined {
  if (array.length === 0) return;
  const totalWeight = array.reduce((w, e) => w + e.weight, 0);
  const targetWeight = totalWeight * getRandom();
  let runningWeight = 0;
  for (const entry of array) {
    if (targetWeight < runningWeight + entry.weight) return entry.item;
    runningWeight += entry.weight;
  }
  return array[0].item;
}

export function getCoordinatesFromRandomValue(
  random: number,
  options: { x: { min: number; max: number }; y: { min: number; max: number } },
): { x: number; y: number } {
  const xRange = options.x.max - options.x.min;
  const xBase = Math.floor(random * xRange);
  const xCoord = options.x.min + xBase;
  const randomRemainder = random * xRange - xBase;
  const yRange = options.y.max - options.y.min;
  const yCoord = options.y.min + Math.floor(randomRemainder * yRange);
  return { x: xCoord, y: yCoord };
}
